@@ -1,19 +1,20 @@
 package Test::Taint;
 
+## no critic (Bangs::ProhibitVagueNames)
+## We're dealing with abstract vars like "$var" in this code.
+
 =head1 NAME
 
 Test::Taint - Tools to test taintedness
 
 =head1 VERSION
 
-Version 1.02
-
-    $Header: /home/cvs/test-taint/Taint.pm,v 1.13 2004/04/23 04:30:59 andy Exp $
+Version 1.06
 
 =cut
 
 use vars qw( $VERSION );
-$VERSION = "1.02";
+$VERSION = "1.06";
 
 =head1 SYNOPSIS
 
@@ -42,7 +43,7 @@ in standard L<Test::More> style.
 use strict;
 use warnings;
 
-use DynaLoader;
+use base 'DynaLoader';
 use Test::Builder;
 use overload;
 use Scalar::Util;
@@ -50,9 +51,15 @@ use vars qw( $TAINT );
 
 my $Test = Test::Builder->new;
 
-use vars qw( @EXPORT @ISA );
-@EXPORT = qw( taint taint_deeply tainted tainted_ok untainted_ok taint_checking taint_checking_ok );
-@ISA = qw(DynaLoader);
+use vars qw( @EXPORT );
+@EXPORT = qw(
+    taint             taint_deeply
+    tainted           tainted_deeply
+    tainted_ok        tainted_ok_deeply
+    untainted_ok      untainted_ok_deeply
+    taint_checking
+    taint_checking_ok
+);
 
 bootstrap Test::Taint $VERSION;
 
@@ -61,11 +68,51 @@ sub import {
     my $caller = caller;
     no strict 'refs';
     for my $sub ( @EXPORT ) {
-        *{$caller.'::'.$sub} = \&$sub;
+        *{$caller.'::'.$sub} = \&{$sub};
     }
     $Test->exported_to($caller);
     $Test->plan(@_);
-}
+} # import
+
+sub _deeply_traverse {
+    my $callback = shift;
+    my @stack    = \@_;
+
+    my %seen;
+
+    while(@stack) {
+        my $node = pop @stack;
+
+        # skip the node if its not a reference
+        next unless defined $node;
+
+        my($realpack, $realtype, $id) = overload::StrVal($node) =~ /\A(?:(.+)\=)?(HASH|ARRAY|GLOB|SCALAR|REF)\((0x[[:xdigit:]]+)\)\z/
+            or next;
+
+        # taint the contents of tied objects
+        if(my $tied = $realtype eq 'HASH'   ? tied %{$node} :
+                      $realtype eq 'ARRAY'  ? tied @{$node} :
+                      $realtype eq 'SCALAR' ? tied ${$node} :
+                      $realtype eq 'REF'    ? tied ${$node} : undef) {
+            push @stack, $tied;
+            next;
+        }
+
+        # prevent circular references from being traversed
+        no warnings 'uninitialized';
+        next if $seen{$realpack, $realtype, $id}++;
+
+        # perform an action on the node, then push them on the stack for traversal
+        push @stack,
+            $realtype eq 'HASH'   ? $callback->(values %{$node}) :
+            $realtype eq 'ARRAY'  ? $callback->(@{$node})        :
+            $realtype eq 'SCALAR' ? $callback->(${$node})        :
+            $realtype eq 'REF'    ? $callback->(${$node})        :
+            map $callback->(*$node{$_}), qw(SCALAR ARRAY HASH);   #must be a GLOB
+    }
+
+    return;
+} # _deeply_traverse
 
 =head1 C<Test::More>-style Functions
 
@@ -87,7 +134,7 @@ sub taint_checking_ok {
     $Test->ok( $ok, $msg );
 
     return $ok;
-} # tainted_ok
+} # taint_checking_ok
 
 =head2 tainted_ok( $var [, $message ] )
 
@@ -123,7 +170,48 @@ sub untainted_ok {
     $Test->ok( $ok, $msg );
 
     return $ok;
-} # tainted_ok
+} # untainted_ok
+
+=head2 tainted_ok_deeply( $var [, $message ] )
+
+Checks that I<$var> is tainted.  If I<$var>
+is a reference, it recursively checks every
+variable to make sure they are all tainted.
+
+    tainted_ok_deeply( \%ENV );
+
+=cut
+
+sub tainted_ok_deeply {
+    my $var = shift;
+    my $msg = shift;
+
+    my $ok = tainted_deeply( $var );
+    $Test->ok( $ok, $msg );
+
+    return $ok;
+} # tainted_ok_deeply
+
+=head2 untainted_ok_deeply( $var [, $message ] )
+
+Checks that I<$var> is not tainted.  If I<$var>
+is a reference, it recursively checks every
+variable to make sure they are all not tainted.
+
+    my %env = my_validate( \%ENV );
+    untainted_ok_deeply( \%env );
+
+=cut
+
+sub untainted_ok_deeply {
+    my $var = shift;
+    my $msg = shift;
+
+    my $ok = !tainted_deeply( $var );
+    $Test->ok( $ok, $msg );
+
+    return $ok;
+} # untainted_ok_deeply
 
 =head1 Helper Functions
 
@@ -139,7 +227,7 @@ Returns true if taint checking is enabled via the -T flag.
 
 sub taint_checking() {
     return tainted( $Test::Taint::TAINT );
-}
+} # taint_checking
 
 =head2 tainted( I<$var> )
 
@@ -148,10 +236,42 @@ Returns boolean saying if C<$var> is tainted.
 =cut
 
 sub tainted {
-    no warnings;
+    no warnings qw(void uninitialized);
 
-    return !eval { join("",@_), kill 0; 1 };
-}
+    return !eval { local $SIG{__DIE__} = 'DEFAULT'; join('', shift), kill 0; 1 };
+} # tainted
+
+=head2 tainted_deeply( I<$var> )
+
+Returns boolean saying if C<$var> is tainted.  If
+C<$var> is a reference it recursively checks every
+variable to make sure they are all tainted.
+
+=cut
+
+sub tainted_deeply {
+  my $is_tainted = 1;
+
+  _deeply_traverse(
+      sub {
+          foreach (@_) {
+            next
+              if not defined
+              or ref
+              or Scalar::Util::readonly $_
+              or tainted $_;
+
+            $is_tainted = 0;
+            last;
+          }
+
+          return @_;
+      },
+      shift,
+  );
+
+  return $is_tainted;
+} # tainted_deeply
 
 =head2 taint( @list )
 
@@ -171,7 +291,7 @@ sub taint {
     for ( @_ ) {
         _taint($_) unless ref or Scalar::Util::readonly $_;
     }
-}
+} # taint
 
 # _taint() is an external function in Taint.xs
 
@@ -187,43 +307,13 @@ the tied object.
 =cut
 
 sub taint_deeply {
-    my @stack = \@_;
-
-    my %seen;
-
-    while(@stack) {
-        my $node = pop @stack;
-
-        # skip the node if its not a reference
-        next unless defined $node;
-
-        my($realpack, $realtype, $id) = overload::StrVal($node) =~ /\A(?:(.+)\=)?(HASH|ARRAY|GLOB|SCALAR|REF)\((0x[[:xdigit:]]+)\)\z/
-            or next;
-
-        # taint the contents of tied objects
-        if(my $tied = $realtype eq 'HASH'   ? tied %$node : 
-                      $realtype eq 'ARRAY'  ? tied @$node :
-                      $realtype eq 'SCALAR' ? tied $$node : undef)  {
-            push @stack, $tied;
-            next;
-        }
-
-        # prevent circular references from being traversed
-        no warnings 'uninitialized';
-        next if $seen{$realpack, $realtype, $id}++;
-
-        # attempt to taint all non-references, then traverse the references
-        push @stack,
-            $realtype eq 'HASH'                         ? do  { taint values %$node; values %$node  }                       :
-            $realtype eq 'ARRAY'                        ? do  { taint @$node;        @$node         }                       :
-            $realtype eq 'GLOB'                         ? map { taint *$node{$_};    *$node{$_}     } qw(SCALAR ARRAY HASH) :
-            $realtype eq 'SCALAR' || $realtype eq 'REF' ? do  { taint $$node;        $$node         }                       :
-            ();
-    }
+    _deeply_traverse(
+        sub { taint @_; @_ },
+        @_,
+    );
 
     return;
-}
-
+} # taint_deeply
 
 BEGIN {
     MAKE_SOME_TAINT: {
@@ -236,7 +326,7 @@ BEGIN {
         last if tainted $TAINT;
 
         # Let's try again. Maybe somebody cleaned those.
-        $TAINT = substr(join("", @ARGV, %ENV), 0, 0);
+        $TAINT = substr(join('', @ARGV, %ENV), 0, 0);
         last if tainted $TAINT;
 
         # If those don't work, go try to open some file from some unsafe
@@ -244,10 +334,11 @@ BEGIN {
         # (Yes, even reading from /dev/null works!)
         local(*FOO);
         for ( qw(/dev/null / . ..), values %INC, $0, $^X ) {
+            next unless defined $_;
             if ( open FOO, $_ ) {
-                my $data;
-                if ( defined sysread FOO, $data, 1 ) {
-                    $TAINT = substr( $data, 0, 0 );
+                my $potentially_tainted_data;
+                if ( defined sysread FOO, $potentially_tainted_data, 1 ) {
+                    $TAINT = substr( $potentially_tainted_data, 0, 0 );
                     last if tainted $TAINT;
                 }
             }
@@ -256,7 +347,7 @@ BEGIN {
     }
 
     # Sanity check
-    die "Our taintbrush should have zero length!" if length $TAINT;
+    die 'Our taintbrush should have zero length!' if length $TAINT;
 }
 
 
